package CVSMonitor::HTML::Table;

# Package to generate large and fairly complex tables in a single hit.

use strict;
use UNIVERSAL 'isa';





#####################################################################
# Defaults

# Default background colours
our $DEFAULT_COLOR = {
	bg     => '#FFFFFF',
	title  => '#CCCCCC',
	total  => '#EEEEEE',
	border => '#999999',
	altrow => '#F4F4F4',
	};





#####################################################################
# Main Methods

sub table {
	my $class = shift;
	my $content = $class->table_rows( shift, @_ ) or return undef;
	$class->table_wrap( $content, @_ );
}

# Generate just the rows portion of the table, without the wrapping
sub table_rows {
	my $class = shift;

	# The data must be an Array of Arrays. We will stringify anything deeper
	my $data = (ref $_[0] eq 'ARRAY' and ! grep { ref $_ ne 'ARRAY' } @{$_[0]} )
		? shift : die( 'Invalid table data structure' );

	# Set colors
	my $opt = @_ ? shift : {};
	ref $opt eq 'HASH' or die('Options not passed as HASH ref');
	$opt->{bgcolor} = $DEFAULT_COLOR->{bg} unless defined $opt->{bgcolor};
	$opt->{titlecolor} = $DEFAULT_COLOR->{title} unless $opt->{titlecolor};
	$opt->{totalcolor} = $DEFAULT_COLOR->{total} unless $opt->{totalcolor};
	if ( $DEFAULT_COLOR->{altrow} and ! $opt->{altrowcolor} ) {
		$opt->{altrowcolor} = $DEFAULT_COLOR->{altrow};
	}

	# Get the depth and width of the table
	$opt->{rows} = $#$data;
	$opt->{columns} = 0;
	foreach ( map { $#$_ } @$data ) {
		$opt->{columns} = $_ if $_ > $opt->{columns};
	}

	# Create the properties map
	my %properties = (
		top => {
			color  => ($opt->{topcolor} || $opt->{titlecolor}),
			align  => ($opt->{topalign} || 'center'),
			valign => $opt->{topvalign},
			bold   => (exists $opt->{boldtop} ? $opt->{boldtop} : 1),
			class  => $opt->{topclass},
			style  => $opt->{topstyle},
			nowrap => $opt->{topnowrap},
			},
		bottom => {
			color  => ($opt->{bottomcolor} || $opt->{totalcolor}),
			align  => ($opt->{bottomalign} || 'right'),
			valign => $opt->{bottomvalign},
			bold   => (exists $opt->{boldbottom} ? $opt->{boldbottom} : 1),
			class  => $opt->{bottomclass},
			style  => $opt->{bottomstyle},
			nowrap => $opt->{bottomnowrap},
			},
		left => {
			color  => ($opt->{leftcolor} || $opt->{titlecolor}),
			align  => ($opt->{leftalign} || 'left'),
			valign => $opt->{leftvalign},
			bold   => (exists $opt->{boldleft} ? $opt->{boldleft} : 1),
			class  => $opt->{leftclass},
			style  => $opt->{leftstyle},
			nowrap => $opt->{leftnowrap},
			},
		right => {
			color  => ($opt->{rightcolor} || $opt->{totalcolor}),
			align  => ($opt->{rightalign} || 'right'),
			valign => $opt->{rightvalign},
			bold   => (exists $opt->{boldright} ? $opt->{boldright} : 1),
			class  => $opt->{rightclass},
			style  => $opt->{rightstyle},
			nowrap => $opt->{rightnowrap},
			},
		altrow => {
			color  => ($opt->{altrowcolor} || $opt->{normalcolor} || $opt->{bgcolor}),
			align  => ($opt->{normalalign} || 'left'),                      # Same as normal
			valign => $opt->{normalvalign},                                 # Same as normal
			bold   => (exists $opt->{boldnormal} ? $opt->{boldnormal} : ''), # Same as normal
			class  => ($opt->{altrowclass} || $opt->{normalclass}),
			style  => ($opt->{altrowstyle} || $opt->{normalstyle}),
			},
		normal => {
			color  => ($opt->{normalcolor} || $opt->{bgcolor}),
			align  => ($opt->{normalalign} || 'left'),
			valign => $opt->{normalvalign},
			bold   => (exists $opt->{boldnormal} ? $opt->{boldnormal} : ''),
			class  => $opt->{normalclass},
			style  => $opt->{normalstyle},
			nowrap => $opt->{normalnowrap},
			},
		);

	# For each of the properties, create it's tag options hash
	foreach my $type ( values %properties ) {
		my $tag_options = {};
		$tag_options->{bgcolor} = $type->{color} if $type->{color};
		$tag_options->{align}   = $type->{align} if defined $type->{align};
		$tag_options->{valign}  = $type->{valign} if defined $type->{valign};
		$tag_options->{class}   = $type->{class} if defined $type->{class};
		$tag_options->{style}   = $type->{style} if defined $type->{style};
		$tag_options->{nowrap}  = undef if ($opt->{nowrap} || $type->{nowrap});
		$type->{tag_options}    = $tag_options;
	}

	# Create the content buffer
	my @content = ();

	# Generate the inside rows/cells
	foreach my $row ( 0 .. $opt->{rows} ) {
		push @content, '  <tr>';
		foreach my $col ( 0 .. $opt->{columns} ) {
			# Establish the type
			my $type = $class->table_cell_type( $row, $col, $opt );
			$type = $properties{$type};

			# Get and cell contents and bold if needed.
			my $cell = $data->[$row]->[$col];
			$cell = defined $cell ? "$cell" : ''; # Stringify objects
			$cell = '&nbsp;' if $cell eq '';
			$cell = "<b>$cell</b>" if $type->{bold};

			# Create the tag cell and add it
			$cell = CVSMonitor::HTML->tag_pair( 'td', $type->{tag_options}, $cell );
			push @content, "    $cell";
		}
		push @content, '  </tr>';
	}

	join "\n", @content;
}

# Wrap a bunch of rows in the double layer funky table
# Returns it as a single string, for convenience
sub table_wrap {
	my $class = shift;
	my $content = shift;
	my $opt = shift || {};

	# Set some defaults
	$opt->{cellpadding} = 3 unless exists $opt->{cellpadding};
	$opt->{innerwidth}  = 1 unless exists $opt->{innerwidth};
	$opt->{outerwidth}  = 1 unless exists $opt->{outerwidth};

	# Set the width of the table
	my $extras = exists $opt->{width} ? qq~ width="$opt->{width}"~ : '';
	$extras .= qq~ align="$opt->{align}"~ if exists $opt->{align};

	# Shortcut for the case where both widths are zero,
	# and this is probably meant to be a layout table.
	if ( $opt->{innerwidth} == 0 and $opt->{outerwidth} == 0 ) {
		return qq~
		<table border="0" cellspacing="0" cellpadding="$opt->{cellpadding}"$extras>
		$content
		</table>
		~;
	}

	# Work out the border colour
	$opt->{bordercolor} = $DEFAULT_COLOR->{border} unless exists $opt->{bordercolor};
	my $border = $opt->{bordercolor} ? qq~ bgcolor="$opt->{bordercolor}"~ : '';
	$border .= qq~ class="$opt->{borderclass}"~ if $opt->{borderclass};

	qq~
	<table border="0" cellspacing="0" cellpadding="$opt->{outerwidth}"$extras>
	<tr><td$border>
	<table border="0" cellspacing="$opt->{innerwidth}" cellpadding="$opt->{cellpadding}" width="100%">
	$content
	</table></td></tr></table>
	~;
}

sub table_cell_type {
	my ($class, $row, $col, $opt) = @_;
	($row == 0 and $opt->{top})                                       ? 'top'
	: ($col == 0 and $opt->{left})                                    ? 'left'
	: ($opt->{rows} and $opt->{bottom} and $row == $opt->{rows})      ? 'bottom'
	: ($opt->{columns} and $opt->{right} and $col == $opt->{columns}) ? 'right'
	: (($row + ($opt->{top} ? 1 : 0)) % 2)                            ? 'altrow'
	:                                                                   'normal';
}





#####################################################################
# Secondary Methods

# When building tables

sub layout {
	my $class = shift;
	my $data = (ref $_[0] eq 'ARRAY') ? shift : return undef;
	my %options = map { lc $_ } @_;

	# How many elements are there?
	my $elements = scalar @$data or return $data;

	# Determine the flow direction
	my $direction = CVSMonitor::Util::string($options{direction}) ? $options{direction} : 'left-top';
	unless ( $direction =~ /^(?:left-top|top-left)$/ ) {
		die "Invalid table layout direction '$direction'";
	}

	# Determine the dimensions of the layout
	my $cols = CVSMonitor::Util::integer( $options{cols} );
	my $rows = CVSMonitor::Util::integer( $options{rows} );
	if ( $cols ) {
		$cols = $elements if $cols > $elements;
		$rows = int($elements / $cols + 0.5);
	} elsif ( $rows ) {
		$rows = $elements if $rows > $elements;
		$cols = int($elements / $rows + 0.5);
	} else {
		# No cols or rows given, the default is two cols
		$cols = 2;
		$rows = int($elements / $cols + 0.5);
	}

	# Split on the layout direction
	my @layout = ();
	if ( $direction eq 'left-top' ) {
		while ( @$data ) {
			my @row = splice( @$data, 0, $cols );
			@row = map { ref $_ eq 'ARRAY' ? @$_ : "$_" } @row;
			push @layout, \@row;
		}
	} elsif ( $direction eq 'top-left' ) {
		@layout = map { [] } (1 .. $rows);
		my $row = 0;
		foreach my $item ( @$data ) {
			push @{$layout[$row++]}, (ref $item eq 'ARRAY' ? @$item : "$item");
			$row = 0 if $row >= $rows;
		}
	}

	\@layout;
}

1;
