package CVSMonitor::MetaData::Module;

# The CVSMonitor::MetaData::Module module provides an object version
# of a CVS module, within a Repository object

use strict;
use UNIVERSAL 'isa';
use base 'CVSMonitor::MetaData::Abstract::Source';
use Cwd        ();
use Clone      ();
use Storable   ();
use File::Spec ();
use Class::Autouse 'File::Flat',
                   'Proc::ProcessTable';





#####################################################################
# Methods that involve contstants

# Get the various full paths for things
sub _checkoutDir    { 'checkout' }
sub pathCheckoutDir { File::Spec->catdir(  $_[0]->root, $_[0]->_checkoutDir ) }
sub pathCvsCacheDir { File::Spec->catdir(  $_[0]->root, 'cvscache' ) }
sub pathLogFile     { File::Spec->catfile( $_[0]->root, 'history.log' ) }
sub pathStructFile  { File::Spec->catfile( $_[0]->root, 'metadata.dat' ) }
sub pathAdminLogFile {
	my $self = shift;
	return undef unless $self->{lock};
	File::Spec->catfile( $self->MetaData->root,
		'adminlogs',
		"$self->{lock}.log" );
}





#####################################################################
# Constructor

sub new {
	my $class = shift;
	my $name = shift or return undef;
	my $module = shift or return undef;
	my $Repository = isa( $_[0], 'CVSMonitor::MetaData::Repository' ) ? shift
		: return $class->_error( "Invalid Repository argument" );
	my $display = shift;

	# Create the new empty object
	my $self = {
		_PARENT     => $Repository,  # Which MetaData does this belong in
		name        => $name,  # Symbolic name
		display     => ($display || $name),  # Display Name
		module      => undef,  # Module name in repository
		state       => "New",  # "New", "Unavailable", "Available", "Transient"
		lock        => undef,  # If defined, pid of process working on module
		_FILES      => undef,  # Files in Module
		_CHANGESETS => undef,  # ChangeSets in Module
		trackers    => 0,      # Links to Tracker type external systems
		};
	bless $self, $class;

	# Purge to remove any old etc files
	unless ( $self->purge ) {
		return $self->_error( "Something was in the way. I tried to remove the obstruction, but failed" );
	}

	$self->{state} = 'Unavailable';
	$self->{module} = $module;

	$self;
}

# Load a module from a cached version in the work area
sub load {
	my $class = shift;
	my $name = shift or return undef;
	my $Repository = shift;
	return undef unless isa( $Repository, 'CVSMonitor::MetaData' );

	# Does the file exist
	my $filename = $Repository->moduleFreezeFile( $name );

	# Check the file
	return $class->class( "Module file does not exist" ) unless -f $filename;

	# In order to avoid nasty destroy warnings during the quit() phase, we need
	# To have all modules loaded that are likely to get ->DESTROY calls. That is,
	# any classes present in the Storable object.
	Class::Autouse->load( 'CVSMonitor::MetaData::File' );
	Class::Autouse->load( 'CVSMonitor::MetaData::Version' );
	Class::Autouse->load( 'CVSMonitor::MetaData::Author' );
	Class::Autouse->load( 'CVSMonitor::MetaData::ChangeSet' );

	# Deserialise the content
	my $self = Storable::retrieve( $filename ) or return undef;
	bless $self, $class;

	# Looks good, connect up the parent
	$self->{_PARENT} = $Repository;

	$self;
}

# Save a module to the work area
sub save {
	my $self = shift;
	my $file = $self->pathStructFile;

	# No need to save if we are already
	return 1 if -e $file;

	# Make sure everything is loaded
	unless ( $self->{_FILES} ) {
		$self->populate or return undef;
	}

	# Temporarily unlink the parent
	my $Repository = $self->{_PARENT};
	$self->{_PARENT} = undef;

	# Store, and relink the parent
	my $rv = Storable::store( $self, $file );
	$self->{_PARENT} = $Repository;
	return $self->_error( "Error storing Module" ) unless $rv;

	1;
}

# Reload this module from a newer version.
# This will only be called from the MetaData->reload method.
sub _reload {
	my $self = shift;
	my $replacement = shift;

	# Clear in preperation
	$self->clear or return undef;

	# Replace all our existing properties, EXCEPT for _PARENT
	# with the replacement's
	foreach ( grep { $_ ne '_PARENT' } keys %$self ) {
		delete $self->{$_};
	}
	foreach ( grep { $_ ne '_PARENT' } keys %$replacement ) {
		$self->{$_} = $replacement->{$_};
	}

	1;
}

# Update our object to reflect that the module itself was deleted
# elsewhere. We convert outselves into an error trap, so that if 
# an existing reference is used, an error is generated.
# The ->delete method is reserved for the user deleting a module.
sub _reloadDelete {
	my $self = shift;

	# Remove all out properties
	foreach ( keys %$self ) {
		delete $self->{$_};
	}

	# Rebless ourself to a CVSMonitor::MetaData::Module::Deleted
	bless $self, 'CVSMonitor::MetaData::Module::Deleted';

	1;
}





#####################################################################
# Access methods

sub shortName   { $_[0]->{name} }
sub display     { $_[0]->{display} }
sub state       { $_[0]->{state} }
sub dateRaw     { $_[0]->{date} }
sub updateTime  { $_[0]->{updatetime} }
sub footprint   { $_[0]->{footprint} }

# Get to our various parents
sub Repository  { $_[0]->{_PARENT} }
sub MetaData    { $_[0]->{_PARENT}->MetaData }

# Get the module's full name
sub name {
	my $self = shift;

	# Get our Repository
	my $Repository = $self->Repository or return undef;

	# Merge the two names
	$Repository->name . '.' . $self->{name};
}

# Pass the date value through to the string generator
sub date { CVSMonitor::MetaData->date( $_[0]->{date} ) }

# Generic getter and setter
sub get { $_[0]->{$_[1]} }
sub set {
	my $self = shift;
	my $field = shift;
	my $value = shift;
	$self->{$field} = $value;
	$self->MetaData->dirty;

	1;
}

# Get the cvsroot and public cvsroot
sub cvsroot       { $_[0]->Repository->cvsroot }
sub cvsrootPublic { $_[0]->Repository->publicCvsroot }

# Get the root directory for this module.
sub root {
	my $self = shift;
	File::Spec->catdir( $self->MetaData->root, $self->name );
}

# Get the various things
# Is the module available for either read or write
sub available {
	my $self = shift;
	my $type = shift;
	$type = 'read' unless (defined $type and $type eq 'write');
	if ( $type eq 'read' ) {
		return 0 if $self->{state} eq 'New';
		return 0 if $self->{state} eq 'Checking Out';
		return 0 if $self->{state} eq 'Unavailable';

		# We can read while updating
		return 1;
	} else {
		return $self->{state} eq 'Available' ? 1 : 0;
	}
}

# Get the size of the log file
sub _getLogFileSize {
	my $self = shift;
	return undef unless $self->available;

	# Get the filename
	my $logfile = $self->pathLogFile;
	return undef unless -e $logfile;
	return undef unless -r $logfile;

	# Get the information about the file
	my @filedata = stat $logfile;

	# Return the size in bytes
	$filedata[7];
}

sub calculateFootprint {
	my $self = shift;
	my $root = $self->root;

	# Many operating systems don't support the -sb option for du to 
	# provide the size in bytes. Instead, we need to execute a -sk to
	# get the size in kilobytes, and then multiply by 1024.

	# But, we live in hope, so we try to do it with -sb first, and see if it
	my $solaris=1;	my $command = "du -sb $root 2>/dev/null";
	my @output = `$command`;
	my $no_sb = @output ? 0 : 1;

	if ( $no_sb ) {
		# Doesn't support -sb, fall back to the -sk version
		$command = "du -sk $root 2>/dev/null";
		@output = `$command`;
	}

	# Process the results
	chomp @output;
	my $total = 0;
	foreach ( @output ) {
		if ( /^du\:/ ) {
			# Message from du, file probably missing. Ignore
			next;
		} elsif ( /^(\d+)/ ) {
			# A value in bytes, add
			$total += $1;
		}
	}

	# If we don't support -sb, multiply to get the full value
	$total = $total * 1024 if $no_sb;
	$self->{footprint} = $total;

	1;
}

# Can the module be classified as "transient".
# That is, it is either updating or checking out.
# Does not check for brokenness. A transient module may be broken.
sub transient {
	my $self = shift;
	return 1 if $self->{state} eq 'Checking Out';
	return 1 if $self->{state} eq 'Updating';
	0;
}

# Could the module be classified as being in an illegal ( broken ) state.
sub broken {
	my $self = shift;
	if ( $self->{lock} ) {
		my $rv = $self->lockHolderExists;
		return undef unless defined $rv;
		return $rv ? 0 : 1;
	} else {
		if ( $self->{state} eq 'Checking Out'
		or $self->{state} eq 'Updating' ) {
			return 1;
		}
	}
}

# Create a stub to freeze into the index
sub _freezable {
	my $self = shift;

	# Copy the main hash, excluding internals
	my $copy = bless {}, ref $self;
	foreach ( keys %$self ) {
		next if /^_/;
		$copy->{$_} = ref $self->{$_}
			? Clone::clone( $self->{$_} ) # Only clone if we need to
			: $self->{$_};
	}

	$copy;
}





#####################################################################
# Locking

# Get a lock on a Module
sub lock {
	my $self = shift;

	# Is there an existing lock?
	if ( $self->{lock} ) {
		# Return true if it is out lock
		return 1 if $self->lockIsOurs;

		# Check to see if the process exists. If it doesn't 
		# we can aquire the lock from it.
		if ( $self->lockHolderExists ) {
			# We can't take the lock
			return $self->_error( "Failed to aquire module lock. Lockholder still exists" );
		} else {
			# Remove the existing lock.
			delete $self->{lock};
		}
	}

	# Create the lock for our process id
	$self->set('lock', $$ );
	1;
}

# Is a module locked
sub locked {
	my $self = shift;
	defined $self->{lock} ? 1 : 0;
}

# Get the PID of the lock hodler
sub lockPid { $_[0]->{lock} }

# Determine if the lockholder exists
sub lockHolderExists {
	my $self = shift;
	return undef unless $self->{lock};

	# Grab the process table and iterate through to find
	# the process for the lockholder.
	my $ProcessTable = Proc::ProcessTable->new();
	my $matched = scalar grep { $_->pid == $self->{lock} } @{$ProcessTable->table};
	$matched ? 1 : 0;
}

# Determine if the lock is outs
sub lockIsOurs {
	my $self = shift;
	return undef unless $self->{lock};
	$$ == $self->{lock} ? 1 : 0;
}


# Release a lock we own.
sub releaseLock {
	my $self = shift;
	return undef unless $self->lockIsOurs;
	delete $self->{lock};
	$self->MetaData->dirty;
	1;
}





#####################################################################
# Tracker system integration

# Clone the hash and return
sub trackers { $_[0]->{trackers} ? Clone::clone( $_[0]->{trackers} ) : 0 }
sub addTracker {
	my $self = shift;
	my $keyword = lc shift;
	my $url = shift or return undef;

	# Check the arguments
	unless ( $keyword =~ /^\w+$/ ) {
		return $self->_error( 'Invalid tracker keyword format' );
	}
	if ( $self->{trackers} and $self->{trackers}->{$keyword} ) {
		return $self->_error( "Tracker with keyword '$keyword' already exists" );
	}

	# Add the tracker
	$self->{trackers} = {} unless $self->{trackers};
	$self->{trackers}->{$keyword} = $url;

	$self->MetaData->dirty;
	1;
}

sub deleteTracker {
	my $self = shift;
	my $keyword = shift;
	return 1 unless $self->{trackers};
	return 1 unless exists $self->{trackers}->{$keyword};
	delete $self->{trackers}->{$keyword};
	$self->MetaData->dirty;
	1;
}







#####################################################################
# CVS Interaction

# Check a module out from CVS.
# Assumes we are logged in as needed.
sub checkout {
	my $self = shift;
	return $self->_error( "Module is locked" ) if $self->{locked};
	unless ( $self->{state} eq 'Unavailable' ) {
		return $self->_error( "Cannot checkout from a '$self->{state}' state" );
	}

	# Make sure we can save the module state changes
	unless ( $self->MetaData->haveLock ) {
		return $self->_error( "We do not have lock on the index file, cannot proceed" );
	}

	# Lock the Module, and save and release the lock on the index.
	$self->lock or return $self->_error( "Failed to lock module" );
	$self->set('state', 'Checking Out');
	$self->MetaData->save or return $self->_error( "Error saving index" );

	# Prepare by establishing some values
	my $name = $self->name;
	my $cvsroot = $self->cvsroot;
	my $dir = Cwd::fastcwd();
	my $root = $self->root;
	my $checkout = $self->_checkoutDir;
	my $logfile = $self->pathLogFile;

	# Make sure the root exists
	unless ( -d $root ) {
		File::Flat->makeDirectory( $root ) or return undef;
	}

	# Do we have an error log we can dump to?
	my $errorlog = shift;
	$errorlog = " 2>>$errorlog" if $errorlog;

	# Do the checkout
	chdir $root or return undef;
	my $rv = system( "cvs -d $cvsroot -Q checkout -r 1.1 -d $checkout \"$self->{module}\" $errorlog" );
	chdir $dir or return undef;
	return $self->_error( "An error occured during the 'cvs checkout' command" ) if $rv;

	# Pull the cvs log and dump to the correct file
	chdir $self->pathCheckoutDir or return undef;
	$rv = system( "cvs -d $cvsroot -Q log > $logfile $errorlog" );
	chdir $dir or return undef;
	return $self->_error( "An error occured while pulling the CVS log: $rv" ) if $rv;

	# Make sure that anyone can do anything to any of the files
	my $command = "chmod -R 777 $root";
	$rv = system( $command );
	return $self->_error( "Error updating the permissions on temp files: '$command' failed" ) if $rv;

	# If the struct file exists for some reason, remove it.
	my $struct = $self->pathStructFile;
	if ( -e $struct ) {
		unlink $struct or return undef;
	}

	# Reload the index now we have finished
	$self->MetaData->reload or return undef;

	# Make sure the module is still in existance
	unless ( $self->MetaData->getModule( $name ) ) {
		return $self->_error( "Module was deleted while we were working" );
	}

	# Is our module still locked by us?
	unless ( $self->lockIsOurs ) {
		return $self->_error( "Someone has taken or removed our lock while we were working" );
	}

	# Set the update date
	$self->set( 'date', time );

	# Populate ( and save ) the data structs
	$self->populate or return undef;

	# Calculate the module footprint
	$self->calculateFootprint();

	# Update the state and release the module lock
	$self->set( 'state', 'Available' );
	$self->releaseLock or return undef;

	# Save ourselves, but hold the lock
	$self->MetaData->save( 'lock' );

	return $self;
}

# Update a module.
sub update {
	my $self = shift;
	return $self->_error( "Module is locked" ) if $self->{locked};
	unless ( $self->{state} eq 'Available' ) {
		return $self->_error( "Cannot checkout from a '$self->{state}' state" );
	}

	# Make sure we can save the module state changes
	unless ( $self->MetaData->haveLock ) {
		return $self->_error( "We do not have lock on the index file, cannot proceed" );
	}

	# Lock the Module, and save and release the lock on the index.
	$self->lock or return $self->_error( "Failed to lock module" );
	$self->set('state', 'Updating');
	$self->MetaData->save or return $self->_error( "Error saving index" );

	# Prepare by getting some values
	my $cvsroot  = $self->cvsroot;
	my $dir      = Cwd::fastcwd();
	my $root     = $self->root;
	my $logfile  = $self->pathLogFile;
	my $checkout = $self->pathCheckoutDir;

	# Start update timer
	my $startTime = time;

        # Do we have an error log we can dump to?
        my $errorlog = shift;
        $errorlog = " 2>>$errorlog" if $errorlog;

	# Call cvs to update the module
	chdir $checkout or return undef;
	my $rv = system( "cvs -d $cvsroot -Q update -d -r 1.1 $errorlog" );
	chdir $dir or return undef;
	return $self->_error( "An error occured while updating from CVS" ) if $rv;

	# Run the update and dump to the correct location
	chdir $checkout or return undef;
	$rv = system( "cvs -d $cvsroot -Q log > $logfile $errorlog" );
	chdir $dir or return undef;
	return $self->_error( "An error occured while pulling the CVS log: $rv" ) if $rv;

	# Make sure that anyone can do anything to any of the files
	my $command = "chmod -R 777 $root";
	$rv = system( $command );
	return $self->_error( "Error updating the permissions on temp files: '$command' failed" ) if $rv;

	# If the struct file exists, remove it.
	my $struct = $self->pathStructFile;
	if ( -e $struct ) {
		unlink $struct or return undef;
	}

	# Reload the index now we have finished
	$self->MetaData->reload or return undef;

	# Make sure the module is still in existance
	unless ( $self->MetaData->getModule( $self->name ) ) {
		return $self->_error( "Module was deleted while we were working" );
	}

	# Is our module still locked by us?
	unless ( $self->lockIsOurs ) {
		return $self->_error( "Someone has taken or removed our lock while we were working" );
	}

	# Set the update date
	$self->set('date', time );

	# Populate ( and save ) the data structs
	$self->populate or return undef;

	# Recalculate the module footprint
	$self->calculateFootprint();

	# End the timer, and save the update time
	$self->set('updatetime', time - $startTime);

	# Update the state and release the module lock
	$self->set('state', 'Available');
	$self->releaseLock or return undef;

	# Save ourselves, but hold the lock
	$self->MetaData->save('lock');

	# Done
	return 1;
}

# Determine the minimum amount of data we will have to download in order
# to do an update. Returns value in bytes, and any repository at localhost
# has a cost of zero.
sub updateCost {
	my $self = shift;
	if ( ! $self->available ) {
	} else {
		return $self->_getLogFileSize;
	}
}

# Release a module ( uncheckout ).
# We don't actually do a cvs release, we just purge, and update the various 
# properties of the module
sub release {
	my $self = shift;

	# Do the purge
	$self->purge or return undef;

	# Set or remove some flags
	delete $self->{date};
	delete $self->{updatetime};
	delete $self->{footprint};
	$self->set('state', 'Unavailable');

	# Save ourselves, but hold the lock
	$self->MetaData->save('lock');

	# Done
	return 1;
}

# Delete ourselves from our parent
sub delete {
	my $self = shift;

	# Pass off to our parent repository
	return $self->Repository->deleteModule( $self->shortName );
}

# Slurp in the log contents
sub readLog {
	my $self = shift;
	my $file = $self->pathLogFile;
	return undef unless -f $file;

	# Pull and return the file contents
	my $contents = File::Flat->slurp( $file );

	# Split and return
	return [ split /(?:\015\012|\015|\012)/, $$contents ];
}

# Execute an arbitrary command for this module, and return the results
# Pass this through to the backend
sub getCVSCommandResults {
	return CVSMonitor::Backend::CCVS->_Module_getCVSCommandResults( $_[0], $_[1] );
}





#####################################################################
# Population and Generation

# Remove all files relating to the Module from disk
sub purge {
	my $self = shift;

	# Remove the directory in the scratch directory
	my $root = $self->root or return undef;
	return 1 unless -d $root;
	return File::Flat->remove( $root ) ? 1 : undef;
}

# Clear out the Files, Versions and ChangeSets, so that they
# will be immediately garbage collected. ( Remove circular references ).
sub clear {
	my $self = shift;
	return 1 unless $self->{_FILES} || $self->{_CHANGESETS};

	# Get the Versions, Files, and ChangeSets
	my $Versions = $self->getVersions or return undef;
	my $Files = $self->getFiles or return undef;
	my $ChangeSets = $self->getChangeSets or return undef;

	# Go through and unlink all of them from their parents
	foreach ( @$Versions, @$Files, @$ChangeSets ) {
		undef $_->{_PARENT};
	}

	# Remove the top level links
	$self->{_FILES} = undef;
	$self->{_CHANGESETS} = undef;

	return 1;
}

# Create all the Files, Versions and ChangeSets
sub populate {
	my $self = shift;

	# Clear any existing objects
	$self->clear or return undef;

	# Load from the frozen data if it exists
	if ( -r $self->pathStructFile ) {
		return $self->populateCached or return undef;
	}

	# Get the log file
	my $contents = $self->readLog or return undef;

	# Chop into sections
	my @fileContents = ();
	my @thisFile = ();
	foreach ( @$contents ) {
		if ( $_ =~ /^\={77}$/ ) {
			push @fileContents, [ @thisFile ];
			@thisFile = ();
		} else {
			push @thisFile, $_;
		}
	}
	push @fileContents, [ @thisFile ] if scalar @thisFile;

	# Create the objects from the fragments
	my @objects = ();
	foreach ( @fileContents ) {
		my $File = CVSMonitor::MetaData::File->new( $_, $self );
		return undef unless defined $File;

		push @objects, $File;
	}

	# Set the File Objects
	$self->{_FILES} = \@objects;

	# Next, do the ChangeSets
	# Get the versions
	my $Versions = $self->getVersions() or return undef;

	# Now pass them to be converted into ChangeSets
	my $ChangeSets = CVSMonitor::MetaData::ChangeSet->findChangeSets( $Versions );
	return undef unless $ChangeSets;

	# Set the changesets
	$self->{_CHANGESETS} = $ChangeSets;

	# Save the Module, so we won't have to do this again
	$self->save;

	return 1;
}

# Load in the File and ChangeSet data from the disk,
# instead of loading it from the log files.
sub populateCached {
	my $self = shift;

	# Make sure the file exists
	my $file = $self->pathStructFile;
	return undef unless -e $file && -r $file;

	# In order to avoid nasty destroy warnings during the quit() phase, we need
	# To have all modules loaded that are likely to get ->DESTROY calls. That is,
	# any classes present in the Storable object.
	Class::Autouse->load( 'CVSMonitor::MetaData::File' );
	Class::Autouse->load( 'CVSMonitor::MetaData::Version' );
	Class::Autouse->load( 'CVSMonitor::MetaData::Author' );
	Class::Autouse->load( 'CVSMonitor::MetaData::ChangeSet' );

	# Load the frozen object
	my $frozen = Storable::retrieve( $file ) or return undef;

	# Set out Files and ChangeSet information to the fronzen one's
	$self->{_FILES} = $frozen->{_FILES};
	$self->{_CHANGESETS} = $frozen->{_CHANGESETS};

	# Iterate through and reset the parent values
	$_->{_PARENT} = $self foreach @{$self->{_FILES}};

	# Done
	return 1;
}




#####################################################################
# Access Methods

# Get all the files
sub getFiles {
	my $self = shift;
	unless ( $self->{_FILES} ) {
		$self->populate or return undef;
	}
	return [ @{ $self->{_FILES} } ];
}

# Find a file with a particular name
sub getFile {
	my $self = shift;

	# Make sure they specified a filename
	my $filename = shift;
	unless ( $filename ) {
		return $self->_error( "You did not specify a file name" );
	}

	# Populate the information if needed
	unless ( $self->{_FILES} ) {
		$self->populate or return undef;
	}

	# Iterate through the files to find it
	foreach my $File ( @{ $self->{_FILES} } ) {
		return $File if $File->filename eq $filename;
	}

	# Not found
	return $self->_error( "File '$filename' not found in Module" );
}

# Get the versions, optionally within a certain date range
sub getVersions {
	my $self = shift;
	my $date = shift;
	unless ( $self->{_FILES} ) {
		$self->populate or return undef;
	}

	# Map to versions
	my @Versions = ();
	foreach my $File ( @{$self->{_FILES}} ) {
		push @Versions, @{ $File->getVersions };
	}

	# Filter to the date if needed
	if ( $date ) {
		@Versions = grep { $_->inDate( $date ) } @Versions;
	}

	return scalar @Versions ? \@Versions : 0;
}

# Find a particular version of a file
sub getVersion {
	my $self = shift;
	my $filename = shift or return undef;
	my $revision = shift or return undef;
	my $File = $self->getFile( $filename ) or return undef;
	return $File->getVersion( $revision );
}

# Get the ChangeSets, optionally within a certain date range
sub getChangeSets {
	my $self = shift;
	my $date = shift;
	unless ( $self->{_CHANGESETS} ) {
		$self->populate or return undef;
	}

	# Filter to date if required
	if ( $date ) {
		my @ChangeSets = grep { $_->inDate( $date ) } @{$self->{_CHANGESETS}};
		return scalar @ChangeSets ? \@ChangeSets : 0;
	} else {
		return [ @{ $self->{_CHANGESETS} } ];
	}
}

# Get a ChangeSet based on it's id
sub getChangeSet {
	my $self = shift;
	my $id = shift or return undef;

	# Iterate through the ChangeSets looking for it
	my $ChangeSets = $self->getChangeSets or return undef;
	foreach my $ChangeSet ( @$ChangeSets ) {
		if ( $ChangeSet->id == $id ) {
			return $ChangeSet;
		}
	}

	# Not found
	return $self->_error( "No ChangeSet with the id $id" );
}

# Get an Activity object for this
sub getActivity { CVSMonitor::MetaData::Activity->new( shift ) }

# Build a Range for the Repository
sub getRange {
	my $self = shift;
	my $type = shift || 'Basics';
	my $arg = shift || 'month';
	my $RangeClass = 'CVSMonitor::Range::' 
		. ($type eq 'Files' ? 'Files' : 'Basics');

	# Do the entire thing to a particular depth
	if ( $arg eq 'year' or $arg eq 'month' ) {
		return $RangeClass->newFor( $self, $arg );
	}

	# Limit to a single year
	my $regex = CVSMonitor::DateDepth->seperate_regex( 'year' ) or return undef;
	if ( $arg =~ $regex ) {
		return $RangeClass->year( $arg );
	}

	# Limit to a single month
	$regex = CVSMonitor::DateDepth->seperate_regex( 'month' ) or return undef;
	if ( $arg =~ $regex ) {
		return $RangeClass->month( $arg );
	}

	return $self->_error( "Unsupported argument" );
}

# Build a populated Range from the Module
sub getRangePopulated {
	my $self = shift;
	my $type = shift;
	my $arg = shift;

	# Get the empty Range
	my $Range = $self->getRange( $type, $arg ) or return undef;

	# Populate the range
	$Range->add( $self ) or return undef;
	return $Range;
}

# Get the Authors in the Module
sub getAuthors {
	my $self = shift;

	# Break the ChangeSets up by author
	my %hash = ();
	foreach ( @{ $self->{_CHANGESETS} } ) {
		my $name = $_->author;
		if ( exists $hash{$name} ) {
			push @{ $hash{$name} }, $_;
		} else {
			$hash{$name} = [ $_ ];
		}
	}

	# Create the Author objects
	my @Authors = ();
	foreach ( sort keys %hash ) {
		my $Author = CVSMonitor::MetaData::Author->new( $hash{$_} );
		return undef unless $Author;
		push @Authors, $Author;
	}

	return \@Authors;
}

# Get a single Author
sub getAuthor {
	my $self = shift;
	my $name = shift;
	return 0 unless $self->containsAuthor( $name );

	# Create the object from ourselves.
	# It can create itself from us quite happily
	return CVSMonitor::MetaData::Author->new( $name, $self );
}

# Does the module contain the author of a particular name
sub containsAuthor {
	my $self = shift;
	my $name = shift or return undef;
	unless ( $self->{_CHANGESETS} ) {
		$self->populate or return undef;
	}

	# Iterate through all the ChangeSets looking for it
	foreach ( @{ $self->{_CHANGESETS} } ) {
		return 1 if $_->author eq $name;
	}

	# Not in here
	return 0;
}





#####################################################################
# Utility methods

sub _error { $CVSMonitor::MetaData::errstr = $_[1]; undef }
sub errstr   { $CVSMonitor::MetaData::errstr }





#####################################################################
package CVSMonitor::MetaData::Module::Deleted;

# This package is used to replace an existing Module with a "deleted"
# message generator when reloading a CVSMonitor::MetaData object in-place.

use strict;

# Catch anything and return an error
sub AUTOLOAD {
	my ($class, $function) = $_[0] =~ m/^(.*)::(.*)$/o;
	return CVSMonitor::MetaData::Module->_error( "Unable to ->$function, the Module has been deleted" );
}

1;

__END__

# Copyright (C) 2002-2004 Adam Kennedy
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Should you wish to utilise this software under a different licence, 
# please contact the author.
